package uk.co.hcsoftware.xmogrify;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import uk.co.hcsoftware.yago.Argument;
import uk.co.hcsoftware.yago.Option;
import uk.co.hcsoftware.yago.parsers.GnuParser;
import uk.co.hcsoftware.yago.util.DefaultArgReaders;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerException;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamResult;
import java.io.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.List;


public class Main {
    public static final Logger log = Logger.getLogger(Main.class.getName());

    public static final Option help = new Option("help",false,"print this usage");
    public static final Option<File> inputFile = new Option<File>("input",false
            ,'i',"inputfile",DefaultArgReaders.createFileReader(true)
            ,"file containing input XML (if omitted use std.in)");

    public static final Option<File> outputFile = new Option<File>("output",false
            ,'o',"outputfile",DefaultArgReaders.createFileReader(false)
            ,"file to which to save output XML (if omitted use std.out)");

    public static final Option<File> stylesheet = new Option<File>("xsl",true,'x',"xsl",
            DefaultArgReaders.createFileReader(true),
            "Stylesheet");

   public static final Option overwrite = new Option("overwrite",false,"Overwrite mode: read args (each is an inputfile). Buffer file in memory, apply transform, overwrite same file [false]");

    public static final Option validate = new Option("validate",false,"Validate XML [false]");
    public static final Option notNamespaceAware = new Option("not_namespace_aware",false,"Make parser be not namespace-aware [false]");
    public static final Option loadExternalDtd = new Option("load_external_DTD",false,"Load external DTD [false]");
    public static final Option loadExternalGenEnt = new Option("load_external_ge",false,"Load external general entities [false]");

    private static Option[] opts = {stylesheet,inputFile,outputFile,overwrite
            ,validate,loadExternalDtd,notNamespaceAware,loadExternalGenEnt //parser features
    };
    private static Argument[] arguments = new Argument[]{};
    GnuParser gp = new GnuParser(opts,arguments);

    private static final String LONG_DESCRIPTION =
            "Commandline wrapper for java's built-in XSL transform capabilities";
    private static final String INVOCATION
            = "xmogrify options [xml input]";

    public static void main(String args[]){
        new Main().run(args);
    }

    private void run(String[] args) {
        try{
            gp.parseOrDie(INVOCATION,LONG_DESCRIPTION,args);
            gp.createExclusiveOptionGroup(inputFile,overwrite);
            gp.createExclusiveOptionGroup(outputFile,overwrite);
            if(help.isSet()){
                System.out.println(gp.getUsage(INVOCATION,LONG_DESCRIPTION));
                return;
            }
            log.info("xmogrify version 0.1");

            if(overwrite.isSet()){
                List<String> inputFiles = gp.getExtraArgs();
                for(String s:inputFiles){
                    File f = new File(s);
                    InputStream is = new FileInputStream(f);
                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    byte[] buf = new byte[1024];
                    int i;
                    while((i=is.read(buf))!=-1){
                        baos.write(buf,0,i);
                    }
                    is.close();
                    OutputStream os = new FileOutputStream(f);
                    log.info("transforming file="+f.getAbsolutePath());
                    doTransform(new ByteArrayInputStream(baos.toByteArray()), os);
                    os.close();
                }
            }
            else{
                InputStream is;
                final OutputStream os;
                if(inputFile.isSet()) is=new FileInputStream(inputFile.getArgValue());
                else is = System.in;

                if(outputFile.isSet()) os = new FileOutputStream(outputFile.getArgValue());
                else os = System.out;

                doTransform(is, os);
                if(inputFile.isSet()) is.close();
                if(outputFile.isSet()){
                    os.close();
                    log.info("output written to "+outputFile.getArgValue().getAbsolutePath());
                }
            }
        } catch (Exception e) {
            log.log(Level.SEVERE,"fatal exception",e);
            e.printStackTrace();
        }
    }

    private void doTransform(InputStream is, OutputStream os) throws ParserConfigurationException, SAXException, TransformerException, IOException {
        FileInputStream fis = new FileInputStream(stylesheet.getArgValue());
        TransformerFactory tf = TransformerFactory.newInstance();
        SAXSource xslSource = new SAXSource(getXmlReader(), new InputSource(fis));
        tf.newTransformer(xslSource).transform(new SAXSource(getXmlReader(),new InputSource(is)), new StreamResult(os));
        fis.close();
    }


    private XMLReader getXmlReader() throws ParserConfigurationException, SAXException {
        SAXParserFactory spf = SAXParserFactory.newInstance();
        if (!loadExternalGenEnt.isSet()) {
            spf.setFeature("http://xml.org/sax/features/external-general-entities",false);
        }
        spf.setValidating(validate.isSet());
        if (!loadExternalDtd.isSet()) {
            spf.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd",false);
        }
        spf.setNamespaceAware(!notNamespaceAware.isSet());
        return spf.newSAXParser().getXMLReader();
    }

}
